#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*========================================

  R E C O V E R _ P R O T _ C U B E S 

  The difference of this function from
  'recover_prot_cubes' with two parameters
  is that it ignores the cubes satisfied
  in the right branch of node 'n'
 
  ASSUMPTIONS:

  1) n.branch == RIGHT

  ==========================================*/
void ExQuant::recover_prot_cubes(Node &n,CUBE &Cubes_to_protect,int confl_jump_depth)
{

  CUBE Red_cubes;


  CUBE *pCubes_to_exclude;
  assert(n.branch == RIGHT);

  if (n.lval == 0)   pCubes_to_exclude = &Flits0[n.var-1];
  else pCubes_to_exclude = &Flits1[n.var-1];
  CUBE &table = hash_table_cubes1.table;
  int &marker = hash_table_cubes1.marker;
  hash_table_cubes1.change_marker();
  hash_table_cubes1.started_using();

  for (uint i=0; i < pCubes_to_exclude->size(); i++)
    table[(*pCubes_to_exclude)[i]] = marker;

  confirm_active_cubes1(Cubes_to_protect,Red_cubes,confl_jump_depth); 
  check_for_new_red_cubes1(Red_cubes);
  hash_table_cubes1.done_using();

} /* end of function recover_prot_cubes */

/*=============================================

  F I N D _ R E S O L V A B L E _ C U B E

  This function returns the index of the cube
  from Cubes that is resolvable with 'C'.

  If such a cube does not exists, it returns -1

  =============================================*/
int ExQuant::find_resolvable_cube(CUBE &C,CUBE &Cubes)
{
  
  for (uint i=0; i < Cubes.size(); i++)
    if (distance(C,F[Cubes[i]]) == 1)
      return(Cubes[i]);

  return(-1);

} /* end of function find_resolvable_cube */

/*================================================

  F O R M _ I N I T _ P R O T _ C U B E S

  This function is called after conflict
  cube 'C' has been added to 'F'.
  This function creates the list of protecting
  cubes for C. 

  For every literal of C it simply finds the
  first cube with such a literal that can 
  be resolved with C

  Initial protection is needed even for a conflict 
  cube that  is currently not a signal one. This cube 
  may become a signal cube later

  ================================================*/
void ExQuant::form_init_prot_cubes()
{

  
  CUBE &C = F.back();
  CUBE &Prot_cubes = Ping.back();
  int cube_ind;

  for (uint i=0; i < C.size(); i++)    {
    int var_ind = abs(C[i])-1;
    if (Var_state[var_ind].xvar == 0) 	{ 
      Prot_cubes.push_back(-1); //skip Y-variables
      continue; 
    }
    if (C[i] < 0)  cube_ind = find_resolvable_cube(C,Flits1[var_ind]);
    else   cube_ind = find_resolvable_cube(C,Flits0[var_ind]);
     
    Prot_cubes.push_back(cube_ind);
  }

} /* end of function form_init_prot_cubes */

/*=================================


  C O M P U T E _ J U M P


  This function is called in the state
  RES_CNFL (when a conflict resolvent 
  is derived).

  It computes the jump level for 
  this conflict as follows.

  Let C = F.back().  First,
  the largest level L where a
  literal of C was assigned
  is computed.

  Then the first decision level
  above  L is obtained.


  ASSUMPTIONS:
  1) F.back() contains the conflict
  cube


  =================================*/
int  ExQuant::compute_jump(void)
{

 
  CUBE &C = F.back();
  if (C.size() == 0) return(-1);
  if (Cube_info[F.size()-1].state != ACTIVE) {
    p();
    cout << "F.back()-> " << F.back() << endl;
    print_Assgns_cut(F.back());
    print_cube_state(F.size()-1);
    exit(ERROR);
  }
  
  // compute the largest assignment level
  
  int max_depth = -1;

  for (uint i=0; i < C.size(); i++) {
    int var_ind = abs(C[i])-1;
    assert(Var_state[var_ind].val < 2);
    if (Var_sel_depth[var_ind] > max_depth)
      max_depth = Var_sel_depth[var_ind];
  }

  assert(max_depth >= 0);
  int jump_depth = fnd_cnfl_jmp_dpth(max_depth);
  return(jump_depth);

} /* end of function compute_jump */



