#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*===================================================

  D E C _ M A K I N G _ S T A T E _ X V A R

  =====================================================*/
AlgState ExQuant::dec_making_state_xvar()
{
  if (verbose)
    printf("dec_making_state_xvar\n");

  if (unit_cube_assgn_xvar())   {
    if (dec_making_flag)
      printf("a unit X-cube is found\n");
    return(BCP);
  }

  

  if (num_sgnl_xvars == 0) {// no variables  to branch on
    if (num_act_sgnl_cubes != 0) {
      p();
      exit(ERROR);
    }
    if (verbose) {
      printf("dec_making: here\n");
      printf("num_assgn_xvars = %d\n",num_assgn_xvars);
    }
    if (num_assgn_xvars > 0)  
      return(BACKTRACKING);
    else return(FINISH); 
  }

  num_xvar_decs++;
 
  int assgn;
  switch (xvar_dec_mode)   {
  case XVAR_MAX_ACT_SGNL_CUBES:
    assgn =  find_sgnl_xvar_assgn(); // assignment selection
    break;
  case XVAR_MAX_ACT_CUBES:
    assgn =  find_xvar_assgn1(); // assignment selection
    break;
  case XVAR_RAND:
    assgn = rand_xvar_assgn();
    break;
  case XVAR_TRIV:
    assgn = triv_xvar_assgn();
    break;
  case XVAR_LIT_ACTIVITY:
    assgn = find_xvar_assgn2();
    if (assgn == 0)
      assgn = find_sgnl_xvar_assgn();
    break;
  default: assert(false);
  }
  
  if (assgn == 0)   {
    p();
    printf("assignment %d is returned\n",assgn);
    printf("num_assgn_yvars = %d\n",num_assgn_yvars);
    printf("num_assgn_xvars = %d\n",num_assgn_xvars);
    printf("num_act_sgnl_cubes = %d\n",num_act_sgnl_cubes);
    printf("num_sgnl_xvars = %d\n",num_sgnl_xvars);
    print_cube_state(10);
    print_ping();
    exit(ERROR);
  }

  if (dec_making_flag) {
    if (assgn < 0)  printf("making  assignment ~x%d\n",-assgn);
    else printf("making assignment x%d\n",assgn);
  }
 

  // initialize queue
 
  Assgn_queue.clear();
  push_assgn(Assgn_queue,-1,assgn);
  bcp_first_node = DEC_LEFT;

  return(BCP);  


} /* end of function dec_making_state_xvar */




/*==============================================

  F O R M _ U N I T _ X C U B E _ Q U E U E

  ===============================================*/
void ExQuant::form_unit_xcube_queue()
{

  Unit_xcubes.clear();
  int assgn;
  for (uint i=0; i < F.size(); i++) {
    if (Cube_info[i].state != ACTIVE) continue;
    if (Cube_info[i].type == Y_CUBE) continue;
    int L = find_length(F[i],assgn);
    if (L == 1)  {
      if (Var_state[abs(assgn)-1].xvar)
        push_assgn(Unit_xcubes,i,assgn);
      else 
        assert(false); // a unit X-cube should not exist 
      // at this point
    }
	 
  }

  //   cout << "formed int_unit_cubes " << int_unit_cubes << endl;

} /* end of function form_unit_xcube_queue */


/*===============================================

  C R E A T E _ I M P L _ V A R _ N O D E

  ================================================*/
void ExQuant::create_impl_var_node( int cube_ind, int assgn)
{
  create_node(cube_ind,assgn);
  Node &n = Tree_path[curr_depth];
  Var_state[n.var-1].val = 2; // reset node variable back to 2
  init_assgn_queue(cube_ind,assgn);
  bcp_first_node = IMPLIED;
} /* end of function create_impl_var_node */

/*===========================================

  R E M O V E _ M A R K E D _ E L E M S

  ===========================================*/
void remove_marked_elems(CUBE &A)
{
  
  int shift = 0;
  for (uint i=0; i < A.size(); i++) {
    if (A[i] == -1) {
      shift++;
      continue;
    }
    if (shift > 0)
      A[i-shift] = A[i];
  }
  assert(shift > 0);
  A.resize(A.size()-shift);

} /* end of function remove_marked_elems */

/*=================================================

  U N I T _ C U B E _ A S S G N _ X V A R

  ================================================*/
bool ExQuant::unit_cube_assgn_xvar()
{
  
  while (Unit_xcubes.size() > 0)  {
    int cube_ind,assgn;
    pop_assgn(Unit_xcubes,cube_ind,assgn); 
    if (Var_state[abs(assgn)-1].val < 2)  continue;       
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    
    create_impl_var_node(cube_ind,assgn);
    return(true);
  }

  int assgn;
  int count = 0;
  for (int i=Last_cnfl_cubes.size()-1; i >= 0; i--)  {
    int cube_ind = Last_cnfl_cubes[i];
    if (Cube_info[cube_ind].state != ACTIVE)  continue;
    if (Cube_info[cube_ind].type == Y_CUBE) continue;
	    
    int L = find_length(F[cube_ind],assgn);
      
    if (L > 1) {
      Last_cnfl_cubes[i] = -1; // mark this element for removal
      count++;
      continue;
    }

    if (L == 1) 
      if (Var_state[abs(assgn)-1].xvar)	  {
        create_impl_var_node(cube_ind,assgn);
        if (count > 0) remove_marked_elems(Last_cnfl_cubes);
        return(true);  
      }  
  }

  if (count > 0) remove_marked_elems(Last_cnfl_cubes);
  return(false);

} /* end of function unit_cube_assgn_xvar */
