#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*===================================================

  D E C _ M A K I N G _ S T A T E _ Y V A R

  =====================================================*/
AlgState ExQuant::dec_making_state_yvar()
{

  if (num_assgn_yvars ==  (int) Yvars.size())     {
    if (verbose) {
      printf("Yvars.size() = %zu\n",Yvars.size());
      printf("num_assgn_yvars = %d\n",num_assgn_yvars);
      printf("num_assgn_xvars = %d\n",num_assgn_xvars);
      print_assgn_vars();
      print_all_var_assgns();
    }
    switch_into_xvar_mode();
    return(DECISION_MAKING);
  }

  if (unit_cube_assgn_yvar())    
    return(BCP);
  
  num_yvar_decs++;  

  int assgn;
  
  assert(num_assgn_yvars < (int) Yvars.size());
  assert(Yas.size() > 0);
  assgn = find_bp_assgn();
 
  
  if (assgn == 0)   {
    p();
    if (Tree_path.size() < 3) print_tree_path();
    printf("num_sgnl_yvars = %d\n",num_sgnl_yvars);
    printf("num_assgn_yvars = %d\n",num_assgn_yvars);
    printf("Yvars.size() = %d\n",(int) Yvars.size());
    cout << "Yvars-> " << Yvars << endl;
    exit(ERROR);
  }

  if (verbose) {
    if (assgn < 0)  printf("%4smaking  assignment ~y%d\n"," ",-assgn);
    else printf("%4smaking assignment y%d\n"," ",assgn);
  }
 
  
  Assgn_queue.clear();
  push_assgn(Assgn_queue,-1,assgn);
  bcp_first_node = DEC_LEFT;
 
  return(BCP); 
} /* end of function dec_making_state_yvar */




/*=================================================

  U N I T _ C U B E _ A S S G N _ Y V A R

  ================================================*/
bool ExQuant::unit_cube_assgn_yvar()
{
  while (Unit_ycubes.size() > 0) {
    int cube_ind,assgn;
    pop_assgn(Unit_ycubes,cube_ind,assgn);
    if (Var_state[abs(assgn)-1].val < 2) continue;
    create_impl_var_node(cube_ind,assgn);
    return(true);
  }

  int assgn;
  int count = 0;
 
  for (int i=Last_cnfl_cubes.size()-1; i >= 0; i--) {
    int cube_ind = Last_cnfl_cubes[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    if (Cube_info[cube_ind].type != Y_CUBE) continue;
    int L = find_length(F[cube_ind],assgn);
    if (L > 1) {
      Last_cnfl_cubes[i] = -1; // mark this element for removal
      count++;
      continue;
    }

    if (L == 1)   {
      create_impl_var_node(cube_ind,assgn);
      if (count > 0)  remove_marked_elems(Last_cnfl_cubes);
      return(true);  
    }  
  }

  if (count > 0) remove_marked_elems(Last_cnfl_cubes);
  return(false);

} /* end of function unit_cube_assgn_yvar */
/*=================================================

  U P D A T E _ L A S T _ C N F L _ C U B E S

  This function is called when 'yvar_mode'
  is set to 'false'.

  This function eliminates from 'last_cnfl_cubes'
  every X-cube (i.e. a cube containig at least
  one X-variable).

  The reason is that if an X-cube C is currently
  unit (due to assignments to Y-variables, if any)
  it will be added to the queue by 
  form_unit_xcube_queue().

  ===================================================*/
void ExQuant::update_last_cnfl_cubes()
{

  int count = 0;
  for (uint i=0; i < Last_cnfl_cubes.size(); i++) {
    int cube_ind = Last_cnfl_cubes[i];
    if (Cube_info[cube_ind].type != Y_CUBE) {
      Last_cnfl_cubes[i] = -1;
      count++;
    }
  }

  if (count > 0)
    remove_marked_elems(Last_cnfl_cubes);

} /* end of function update_last_cnfl_cubes */
/*==============================================

  F O R M _ U N I T _ Y C U B E _ Q U E U E

  ==============================================*/
void ExQuant::form_unit_ycube_queue()
{

  Unit_ycubes.clear();
  int assgn;
  for (uint i=0; i < F.size(); i++) {
    if (Cube_info[i].state != ACTIVE) continue;
    if (Cube_info[i].type != Y_CUBE) continue;
    int L = find_length(F[i],assgn);
    if (L == 1)   {
      assert (Var_state[abs(assgn)-1].xvar == 0);
      push_assgn(Unit_ycubes,i,assgn);
    }	 
  }

} /* end of function form_unit_ycube_queue */


/*===================================================

  S W I T C H _ I N T O  _ X V A R _ M O D E

  =======================================================*/
void ExQuant::switch_into_xvar_mode()
{
  yvar_mode = false;
  form_unit_xcube_queue();
  update_last_cnfl_cubes();
} /* end of function switch_into_xvar_mode */


