#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <time.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

//
//   FIND_MAX_PROPAG
//
int ExQuant::find_max_propag()
{


  int cube_ind = find_act_sgnl_cube();
  assert(cube_ind >= 0);
  CUBE &C = F[cube_ind];

  int max_count = 0;
  int best_ind = -1;
  
  for (uint i=0; i <  C.size(); i++)    {
    int var_ind = abs(C[i]) - 1;
    if (Var_state[var_ind].val < 2) continue;
    if (red_xvar(var_ind)) continue;
    if (Xvar[var_ind] == 0) continue;
    if (C[i] < 0) {
      if (max_count < (int) Flits1[var_ind].size()) {
        max_count = Flits1[var_ind].size();
        best_ind = i;
      }
    }
    else // C[i] > 0
      if (max_count < (int) Flits0[var_ind].size()) {
        max_count = Flits0[var_ind].size();
        best_ind = i;
      }      
  }

  if (best_ind < 0) {
    p();
    print_cube1(C);
    print_Assgns_cut(C);
    exit(ERROR);
  }
  return(-C[best_ind]);

} /* find_max_propag */

//
//  FIND_ACT_SGNL_CUBE
//
int ExQuant::find_act_sgnl_cube()
{
  int cube_ind;
  for (uint i=0; i <  Sgnl_xvars.size (); i++)    {
    int var_ind = Sgnl_xvars[i] - 1;
    if (Var_state[var_ind].val < 2) continue;
    if (red_xvar(var_ind)) continue;
    cube_ind  = check_cubes(Flits0[var_ind]);
    if (cube_ind >= 0) return(cube_ind);
    cube_ind = check_cubes(Flits1[var_ind]);
    if (cube_ind >= 0) return(cube_ind);
  }

  assert(false);
} /* find_act_sgnl_cube */


//
//   CHECK_CUBES
//
int ExQuant::check_cubes(CUBE &Cubes)
{

  for (uint i=0; i < Cubes.size(); i++) {
    int cube_ind = Cubes[i];
    CubeInfo &El = Cube_info[cube_ind];
    if (El.state != ACTIVE) continue;
    if (El.signal) return(cube_ind);
  }

  return(-1);
} /* check_cubes */

/*============================================

  F I N D _ S G N L _ X V A R _ A S S G N
        
  ==========================================*/
int ExQuant::find_sgnl_xvar_assgn()
{

  int max_val = -1;
  int best_var_ind = -1;
  int max_count = -1;

  if (num_act_sgnl_cubes == 1)
    return(find_max_propag());
  
  for (uint i=0; i <  Sgnl_xvars.size (); i++)    {
    int var_ind = Sgnl_xvars[i] - 1;
    if (Var_state[var_ind].val < 2) continue;
    if (red_xvar(var_ind)) continue;
    int count0 = Act_sgnl_cubes_cnt0[var_ind];
    int count1 = Act_sgnl_cubes_cnt1[var_ind];
    
    int val = count0*count1 + count0+count1;
    if (max_val < val)  {
      max_val = val;
      best_var_ind = var_ind;
    }

    if (count0 > max_count) max_count = count0;
    if (count1 > max_count) max_count = count1;
  }
  
  if (best_var_ind == -1) return(0);

  if (Act_sgnl_cubes_cnt0[best_var_ind] > 
      Act_sgnl_cubes_cnt1[best_var_ind])
    return(best_var_ind + 1);
  else 
    return(-(best_var_ind+1));
 
  
} /* end of function find_sgnl_xvar_assgn */

//
//  FIND_BP_ASSGN
//
int ExQuant::find_bp_assgn()
{

  while (Yas.size() > 0) {
    int assgn = Yas.front();
    Yas.pop_front();
    if (Var_state[abs(assgn)-1].val == 2)
      return(-assgn);
  }
  p();
  exit(ERROR);
} /* find_bp_assgn */

/*=======================================

  R A N D _ X V A R _ A S S G N
        
  =========================================*/
int ExQuant::rand_xvar_assgn()
{

  
  int ntries;

  int num_free_xvars = Xvars.size() - num_assgn_xvars;
  ntries = 1 + (lrand48() % num_free_xvars);
  int count = 0;
  for (uint i=0; i <  Xvars.size (); i++)    {
    int var_ind = Xvars[i] - 1;
    if (Var_state[var_ind].val < 2) continue;
    if (++count == ntries) {
      if (lrand48() % 2 == 0) return(-(var_ind+1));
      else return(var_ind+1);
    }
  }
  printf("count = %d\n",count);
  assert(false);
} /* end of function rand_xvar_assgn */
