#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"


/*===================================

  P R I N T _ S G N L  _ S T A T U S

  =================================*/
void ExQuant::print_sgnl_status(int cube_ind,bool depth_flag,int shift)
{

  for (int i=0; i < shift; i++)
    printf(" ");
  
  printf("Cube_info[%d].signal = %d\n",cube_ind,Cube_info[cube_ind].signal);
  
  if (Cube_info[cube_ind].signal && depth_flag)   {
    for (int i=0; i < shift; i++)
      printf(" ");
    printf("Cube_info[%d].depth=%d\n",cube_ind,Cube_info[cube_ind].depth);
  }
 
} /* end of function print_sgnl_status */


/*=========================

  P R I N T _ D S E Q 2

  ========================*/
void ExQuant::print_dseq2(CUBE &S, int cube_ind)
{ 

  printf("(");
  for (uint i=0; i < S.size(); i++)    {
    int var_ind = abs(S[i])-1;
    if (S[i] < 0) printf("~");
    if (Var_state[var_ind].xvar) printf("x%d",var_ind+1);
    else printf("y%d",var_ind+1);
    if (i !=  S.size()-1) printf(" ");
  }
  printf(")");
  printf(" -> {%d}\n",cube_ind);
  
} /* end of function print_dseq2 */


/*=====================================

  P R I N T _ D S E Q 3

  ======================================*/
void ExQuant::print_dseq3(int cube_ind)
{ 
  if (Cube_info[cube_ind].state != REDUNDANT) return;

  print_dseq2(Act_dseqs[cube_ind],cube_ind);
  
} /* end of function print_dseq3 */


/*=======================================

  P R I N T _ S G N L _ C U B E S

  ========================================*/
void ExQuant::print_sgnl_cubes(bool active)
{
  if (active) 
    printf("Current active signal cubes: \n");
  else
    printf("Current signal cubes: \n");
  for (uint i=0; i < F.size(); i++) {
    CubeInfo &El = Cube_info[i];
    if (El.signal == 0) continue;
    if (active && (El.state != ACTIVE)) continue;
    printf("F[%d]: ",i);
    print_Assgns_cut2(F[i]);
    print_cube_state(i);
  }


} /* end of function print_sgnl_cubes */

/*=============================================

  P R I N T _ A S S G N S _ C U T 2

  ==============================================*/
void  ExQuant::print_Assgns_cut2(CUBE &C)
{

  for (uint i=0; i < C.size();i++)    {
    int lit = C[i];
    int var_ind = abs(lit)-1;
    if (i > 0) printf(" ");
    if (lit < 0) printf("~");
    if (Var_state[var_ind].xvar) printf("x%d",var_ind+1);
    else printf("y%d",var_ind+1);

    if (Var_state[var_ind].val == 2) continue;
    int val=0;
    if (lit > 0) val = 1;
    if (val != value(var_ind))
      printf("(s,%d)",Var_sel_depth[var_ind]);
    else
      printf("(f,%d)",Var_sel_depth[var_ind]);
  }
  printf("\n");

} /* end of function print_Assgns_cut2 */

/*======================================

  P R I N T _ V A R _ S T A T E 1

  =====================================*/
void ExQuant::print_var_state1(int var)
{
  print_var_state(var);
  printf("\n");
} /* end of function print_var_state1 */

/*=========================================

  P R I N T _ V A R _ S T A T E 

  ========================================*/
void ExQuant::print_var_state(int var)
{
  int var_ind = var-1;

  if (Var_state[var_ind].xvar) printf("(x%d ",var);
  else printf("(y%d ",var);
  if (act_sgnl_cubes_cnt(var_ind) == 0)  printf("red ");
  else  printf("non-red ");
    
  
  if (Var_state[var_ind].val < 2)    {
    printf("depth=%d val=%d ",Var_sel_depth[var_ind],value(var_ind));
    if (branch_type(var_ind) == RIGHT) 	{
      printf("br=RIGHT");
      if (falsified_branch(var_ind,LEFT)) printf("(l-false,");
      else printf("(l-true,");
      if (falsified_branch(var_ind,RIGHT))
        printf("r-false) ");
      else printf("r-true) ");
    }	     
    else   {
      printf("br=LEFT");
      if (falsified_branch(var_ind,LEFT))
        printf(" (false) ");
      else printf(" (true) ");
    }
  }
  else printf("unass ");
  
  if (Var_state[var_ind].xvar == 1) printf("xvar)  ");
  else printf("yvar)  ");

}  /* end of function print_var_state */
