#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*========================================

  P R I N T _ A L L _ V A R _ A S S G N S

  =========================================*/
void ExQuant::print_all_var_assgns()
{

  CUBE K;
  for (int i=0; i < num_vars; i++)   {
    if (missing_var(i)) continue;
    K.push_back(i+1);
  }

  print_Assgns_cut(K);
} /* end of function print_all_var_assgns */


//
//  PRINT_ASSGN_VARS
//
void ExQuant::print_assgn_vars()
{

  int count = 0;
  for (int i=0; i < num_vars; i++) {
    VarState &Var = Var_state[i];
    if (Var.val == 2) continue;
    if (count++ > 0) printf(" ");
    print_var(i);
  }

  printf("\n");

} /* print_assgn_vars */

//
//  PRINT_VAR
//
void ExQuant::print_var(int var_ind,bool neg_flag)
{

  assert(var_ind < num_vars);
  VarState &Var = Var_state[var_ind];
  if (neg_flag)
    if (Var.val == 0) printf("~");
  if (Var.xvar) printf("x%d",var_ind+1);
  else printf("y%d",var_ind+1);

} /* print_var */

/*================================================

  R E A L _ N U M _ A S S G N _ I N T _ V A R S

  ===============================================*/
int ExQuant::real_num_assgn_int_vars()
{

  int count = 0;
  for (uint i=0; i < Xvars.size();i++)
    if (Var_state[Xvars[i]-1].val < 2) count++;


  return(count);
} /* end of function real_num_assgn_int_vars */

/*========================================
    
  P R I N T _ A L G _ S T A T E 0
      
  This function prints the algorithm 
  state (in short form)

  =======================================*/
void ExQuant::print_alg_state0(AlgState astate)
{

  switch (astate) {
  case DECISION_MAKING:
    printf("DECISION_MAKING");
    break; 
  case  BCP:
    printf("BCP");
    break;
  case BCP_CNFL:
    printf("BCP_CNFL");
    break;
  case BPE:
    printf("BPE");
    break;
  case BACKTRACKING:
    printf("BACKTRACKING "); 
    break;
  case FINISH:
    printf("FINISH");
    break;
  case RESOURCE_EXCEEDED:
    printf("RESOURCE_EXCEEDED");
    break;
  default:
    p();
    printf("print_alg_state:: unknown state!!\n");
    exit(ERROR);
  }
  printf("\n");

} /* end of function print_alg_state0 */

/*========================================
    
  P R I N T _ A L G _ S T A T E
      
  This function prints the algorithm 
  state

  =======================================*/
void ExQuant::print_alg_state(bool flag)
{
  
  
  Node *pn;
  
  if (curr_depth >= 0) pn = &Tree_path[curr_depth];
  else pn = NULL;
  if (flag)  printf("----------------------\n");
  printf("State: ");   
  switch (alg_state)    {
  case DECISION_MAKING:
    if (yvar_mode) printf("Yvar ");
    else printf("Xvar ");
    printf("DECISION_MAKING at depth %d",curr_depth);     
    break; 
  case  BCP:
    printf("BCP at depth %d",curr_depth);
    break;
  case BCP_CNFL:
    printf("BCP_CNFL at depth %d",curr_depth);
    break;
  case BPE:
    assert(curr_depth >= 0);
    printf("BPE");
    if (Var_state[pn->var-1].xvar) 
      printf(" for variable x%d(%d,%d)",pn->var,pn->lval,!pn->lval);
    else 
      printf(" for variable y%d(%d,%d)",pn->var,pn->lval,!pn->lval);
    printf(" at depth %d",curr_depth);
    printf("       (left_ecube = %d, right_ecube = %d)",pn->left_ecube,pn->right_ecube);
    break;
  case BACKTRACKING:
    assert(curr_depth >= 0);
    printf("BACKTRACKING "); 
    if (pn->branch == LEFT)  
      if (Var_state[pn->var-1].xvar)
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~x%d,LEFT) at depth %d",pn->var,curr_depth);
        else
          printf("to node n(x%d,LEFT) at depth %d",pn->var,curr_depth);
      else // Y-variable
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~y%d,LEFT) at depth %d",pn->var,curr_depth);
        else 
          printf("to node n(y%d,LEFT) at depth %d",pn->var,curr_depth);
    else //right branch
      if (Var_state[pn->var-1].xvar)
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~x%d,RIGHT) at depth %d",pn->var,curr_depth);
        else
          printf("to node n(x%d,RIGHT) at depth %d",pn->var,curr_depth);
      else 
        if (Var_state[pn->var-1].val == 0)
          printf("to node n(~y%d,RIGHT) at depth %d",pn->var,curr_depth);
        else 
          printf("to node n(y%d,RIGHT) at depth %d",pn->var,curr_depth);
    break;
  case FINISH:
    printf("FINISH");
    break;
  case RESOURCE_EXCEEDED:
    printf("RESOURCE_EXCEEDED");
    break;
  default:
    p();
    printf("print_alg_state:: unknown state!!\n");
    exit(ERROR);
  }
  printf("\n");

} /* end of function print_alg_state */
