#define  p() \
  printf("%s, line %d, ",__FILE__,__LINE__); \
  printf("assertion failure\n"); \
  printf("gcount = %ld\n",gcount);

//
//  MLIT_TO_LIT
//
inline int mlit_to_lit(Msolver *Slvr,Mlit L)
{
  assert(var(L) < Slvr->nVars());
  int lit = var(L)+1;
  if (sign(L)) lit = -lit;
  return(lit);
} /*mlit_to_lit */

/*================================

       M I S S I N G _ V A R

ASSUMPTIONS:

   1) var_ind < Flits0.size()
      
      var_ind < Flits1.size()

 ================================*/
inline bool ExQuant::missing_var(int var_ind)
{

  if (Flits0[var_ind].size() != 0) return(false);
  if (Flits1[var_ind].size() != 0) return(false);
  return(true);

} /* end of function missing_var */
/*=====================================

    A L R E A D Y _ P R O T E C T E D

 ======================================*/
inline bool ExQuant::already_protected(int cube_ind,int lit_ind)
{
  if (Ping[cube_ind][lit_ind] == -1) return(false);
  return(Cube_info[Ping[cube_ind][lit_ind]].state == ACTIVE);
} /* end of function already_protected */


/*===================================

     C U R R _ N O D E

 ====================================*/
inline Node &ExQuant::curr_node()
{
 assert(curr_depth >= 0);
 return(Tree_path[curr_depth]);
} /* end of function curr_node */


/*========================

   B R A N C H _ T Y P E

=========================*/
inline BranchType ExQuant::branch_type(int var_ind)
{

  return(Tree_path[Var_sel_depth[var_ind]].branch);

} /* end of function branch_type */

/*=================================

  F A L S I F I E D _ B R A N C H

==================================*/
inline bool ExQuant::falsified_branch(int var_ind,BranchType branch)
{

  if (branch == LEFT)
     return(Tree_path[Var_sel_depth[var_ind]].left_ecube >= 0);
  else 
    return(Tree_path[Var_sel_depth[var_ind]].right_ecube >= 0);
} /* end of function branch_type */

/*========================

    C U R R _ V A R

=========================*/
inline int ExQuant::curr_var()
{

  assert(curr_depth >= 0);
  return(Tree_path[curr_depth].var);

} /* end of function curr_var */

/*======================

    V A R _ N O D E

 =====================*/
inline Node& ExQuant::var_node(int var_ind)
{
  assert(Var_state[var_ind].val < 2);
  return(Tree_path[Var_sel_depth[var_ind]]);

} /* end of function var_node */

/*===========================

    R E D _ X V A R

  =========================*/
inline bool ExQuant::red_xvar(int var_ind)
{
  return((Act_sgnl_cubes_cnt0[var_ind] == 0) && 
         (Act_sgnl_cubes_cnt1[var_ind] == 0));

}/* end of function red_xvar */


/*====================================

   X V A R _ C H E C K
 
====================================*/
inline int ExQuant::xvar_check(Node &n)
{
  return(Var_state[n.var-1].xvar);
} /* end of function xvar_check */
/*====================================

   X V A R _ C H E C K
 
====================================*/
inline int ExQuant::xvar_check(int var_ind)
{ 
  return(Var_state[var_ind].xvar);
} /* end of function xvar_check */

/*=========================

      A S S I G N E D

  ========================*/
inline int ExQuant::assigned(int var_ind)
{
  return(Var_state[var_ind].val < 2);
  
} /* end of function assigned */


/*====================================

       V A L U E

 This function returns the value
 of variable 'var_ind+1'

 
====================================*/
inline int ExQuant::value(int var_ind)
{
  return(Var_state[var_ind].val);
  
} /* end of function value */



/*=================================================================

         C U B E _ T Y P E _ C H E C K

   This function returns 
    NONE if C.size() == 0
    X_CUBE if C contains at least one X-variable
    Y_CUBE  if  C  contains only Y-variables
   
  
==================================================================*/
inline CubeType ExQuant::cube_type_check(CUBE &C)
{
 
  if (C.size() == 0) return(NONE);

  for (uint i=0; i < C.size(); i++)
    if (Var_state[abs(C[i])-1].xvar) return(X_CUBE);

  return(Y_CUBE);
} /* end of function cube_type_check */



/*=======================================

      C O N T A I N S _ L I T

  This function returns 'true' if
  cube C contains literal 'lit'.
  Otherwise, it returns 'false'

========================================*/
inline bool contains_lit(CUBE &C, int lit)
{

  for (uint i=0; i < C.size(); i++)
    if (C[i] == lit)
	return(true);
    
  return(false);
} /* end of function contains_lit */

/*=======================================

      C O N T A I N S _ V A R

  This function returns 'true' if
  cube C contains variable 'var'.
  Otherwise, it returns 'false'

========================================*/
inline bool contains_var(CUBE &C, int var)
{

  for (uint i=0; i < C.size(); i++)
    if (abs(C[i]) == var)
	return(true);
    
  return(false);
} /* end of function contains_var */


/*==========================================

      I N S E R T _ I N T O _ T A B L E

  ==========================================*/
inline void ExQuant::insert_into_table(int var_ind,int lit)
{ CUBE &table = hash_table_lits.table;
  int marker = hash_table_lits.marker;

  if (lit < 0)
    table[var_ind] = marker;
  else
    table[var_ind+num_vars] = marker;


} /* end of function instert_into_table */

/*=============================================

         P R E S E N T _ I N _ T A B L E

  ============================================*/

inline bool ExQuant::present_in_table(int var_ind, int lit)
{ CUBE &table = hash_table_lits.table;
  int marker = hash_table_lits.marker;

  if (lit < 0) 
    return(table[var_ind] == marker);
  else
    return(table[var_ind+num_vars] == marker);


  } /* end of function present_in_table */


/*========================================================

  I N I T _ Q U E U E _ I M P L _ A S S G N

  This function is called when current assignment
  to n.var is equal to n.lval

  ========================================================*/
inline void ExQuant::init_queue_impl_assgn(Node &n) 
{
  int val = Var_state[n.var-1].val;
  assert(val == n.lval);
  if (val == 0) init_assgn_queue(n.left_ecube,n.var);
  else init_assgn_queue(n.left_ecube,-n.var);
} /* end of function init_queue_impl_assgn */


/*========================================================

  I N I T _ Q U E U E _ D E C _ A S S G N

  This function is called in the left branch
  to flip the assignment.

  
  ========================================================*/
inline void ExQuant::init_queue_dec_assgn(Node &n) 
{
  int val = Var_state[n.var-1].val;
  assert(val == n.lval);
  if (val == 0) init_assgn_queue(-1,n.var);
  else init_assgn_queue(-1,-n.var);
} /* end of function init_queue_dec_assgn */



/*====================================
     
         F I N D _ L E N G T H

    This function return the current
    length of cube C (the number of 
    unassigned variables of C).

    ATTENTION!!!
    If length of C is greater than 2
    this functions always returns 2.

   ASSUMPTIONS: 
     1) cube C is unsatisfied
======================================*/
inline int ExQuant::find_length(CUBE &C,int &assgn)
{
  int length = 0;
  for (uint i=0; i < C.size();i++)
    if (Var_state[abs(C[i])-1].val==2) {
      if (length++ == 1) // there are at least two unassigned literals 
        return(2);
      assgn = -C[i];
      continue;
    }
  return(length);	  
} /* end of function find_length */




/*=====================================

           X C U B E

  This function returns 'true' if
  cube C contains an X-variable.
  Otherwise, the function returns
  'false' 

  =====================================*/
inline bool ExQuant::Xcube(CUBE &C)
{
  for (uint i=0; i < C.size(); i++) {
    int var_ind = abs(C[i])-1;
    if (Var_state[var_ind].xvar == 1)
      return(true);
  }
 return(false);
} /* end of function Xcube */


/*======================================
 
      P O P _ A S S G N

=======================================*/
inline void pop_assgn(CUBE &queue,int &unit_cube_num,int &assgn)
{
 assgn = queue.back();  // extract assignment
 queue.pop_back();
   
 unit_cube_num = queue.back(); // extract the number of the cube from the assignment is deduced
 queue.pop_back();
} /* end of function pop_assgn */


/*===============================

     P U S H _ A S S G N

  ============================*/
inline void push_assgn(CUBE &queue,int cube_num, int assgn)
{

  queue.push_back(cube_num);
  queue.push_back(assgn);

} /* end of function push_assgn */

/*===================================

     I N I T _ A S S G N _ Q U E U E

  ==================================*/
inline void ExQuant::init_assgn_queue(int cube_num, int assgn)
{

  Assgn_queue.clear();
  push_assgn(Assgn_queue,cube_num,assgn);

} /* end of function init_assgn_queue */
