#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh" 
#include "einline.hh"



/*==============================================

  U P D A T E  _ S A T _ Y C U B E S

  This function turns into ACTIVE 
  satisfied ycubes listed in Cubes

 
  ASSUMPTIONS:
  1) CUBE &Cubes = New_sat_cubes[curr_depth]
  2) This function is called when flipping the
  left branch of the current node. So
  every satisfied signal cube is listed in 'Cubes'
      

  ===============================================*/
void ExQuant::update_sat_ycubes(CUBE &Cubes)
{


  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].type != Y_CUBE) continue;
    if (Cube_info[cube_ind].state != SATISFIED)	{
      p();
      print_cube_state(cube_ind);
      exit(ERROR);
    }
    Cube_info[cube_ind].state = ACTIVE;
  }

  Cubes.clear();
} /* end of function update_sat_ycubes */



/*========================================

  F N D _ O N E _ P I N G _ C U B E 1

  This functions returns the index of
  cube from 'Cube_inds' that
  a) can be resolved with 'C',
  b) is either currently ACTIVE or
  c) becomes active starting with depth 'depth'

  (So in contrast to 'find_one_ping_cube'
  a protecting cube may be currently inactive)

  If no such cube exists, the function
  returns '-1'. In this case, the function
  also computes the maximum depth up to 
  which cube 'C' remains redundant

  =========================================*/
int  ExQuant::fnd_one_ping_cube1(CUBE &C,CUBE &Cube_inds,int depth,int &max_depth)
{

 
  max_depth = -1;
  int dist;
  for (uint i=0; i < Cube_inds.size(); i++)    {
    int cube_ind = Cube_inds[i];
     
    if ((Cube_info[cube_ind].state != ACTIVE) && (Red_cube_depth[cube_ind] < depth))	{
      if (Red_cube_depth[cube_ind] > max_depth)
        max_depth = Red_cube_depth[cube_ind];
      continue;
    }
     
     
    dist = distance(C,F[cube_ind]); 
    assert(dist > 0);
    if (dist == 1) return(cube_ind);
  }

  return(-1);

} /* end of function fnd_one_ping_cube1 */

/*===============================================

  R E C O V E R _ P R O T E C T I O N

  This function checks that  every
  literal of cube F[cube_ind] is protected
  by a cube that is either currently active
  or stays redundant only up to depth 'depth'.
  If this is the case, then 'recover_protection'
  returns 'true'.


  Otherwise, this function returns 'false'.
  In this case, it also computes 
  a) the minimum depth 'min_depth' up to which
  F[cube_ind] remains redundant
  b) The literal due to which 
  F[cube_ind] remain redundant up to 
  depth 'min_depth'

  NO_ASSUMPTIONS:
  1) We DO NOT ASSUME  that if a cube 'cube_ind1'
  of Ping[cube_ind] is ACTIVE, then 'cube_ind'
  is in the array Ped[cube_ind1]

  ===============================================*/
bool ExQuant::recover_protection(int cube_ind,int depth,int &min_depth_lit,int &min_depth)
{
 
 
  if (Cube_info[cube_ind].signal == 0) return(true); // only signal cubes need protection

  CUBE &C = F[cube_ind];

  CUBE &Ping_cubes = Ping[cube_ind];
  if (C.size() != Ping_cubes.size())     {
    p();
    printf("Wrong Ping[%d] for cube F[%d]\n",cube_ind,cube_ind);
    cout << "C-> " << C << endl;
    cout << "Ping_cubes-> " << Ping_cubes << endl;
    exit(ERROR);
  }
 
  min_depth = num_vars+1;
  bool protected_flag = true;
  
  for (uint i=0; i < C.size(); i++)    {
    int var_ind = abs(C[i])-1;
    if (Var_state[var_ind].xvar == 0) continue; // no need to protect Y-literals
    if (Var_state[var_ind].val < 2) continue; // assigned literals cannot be protected
    int cube_ind1 = Ping_cubes[i];
    if (cube_ind1 == -1) { // this literal cannot be protected globally
      min_depth = -1;
      min_depth_lit = C[i];
      protected_flag = false;
      break;
    }
   
    CUBE *pCubes;
    if (C[i] < 0) pCubes = &Flits1[var_ind];
    else pCubes = &Flits0[var_ind];
    int max_depth;
    int ping_cube_ind = fnd_one_ping_cube1(C,*pCubes,depth,max_depth);
 
    if (ping_cube_ind >= 0) 	{	 
      update_prot_cube_info(cube_ind,i,ping_cube_ind);
      continue;
    }

    protected_flag = false;
      

    if (max_depth < min_depth)	{
      min_depth = max_depth;
      min_depth_lit = C[i];
    }    
  }

  return(protected_flag);
} /* end of function recover_protection */

/*================================================

  C O N F I R M _ A C T I V E _ C U B E S


  This function update information regarding
  protecting cubes for the cubes listed 
  in Cubes.

  ASSUMPTIONS:

  1) Cubes of 'Cubes' are already marked as ACTIVE

  2) Red_cubes is empty

  ===============================================*/
void ExQuant::confirm_active_cubes(CUBE &Cubes,CUBE &Red_cubes,int depth)
{
  
  
 
  for (uint i=0; i < Cubes.size(); i++)    {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    if (Cube_info[cube_ind].signal == 0) continue;

    int min_depth;
    int min_depth_lit;
     
    bool success = recover_protection(cube_ind,depth,min_depth_lit,min_depth);
      
      
    if (success) continue;

    CUBE S;
    	    

    seq_gen_red_cube1(S,min_depth_lit);
    num_frw_blocked_cube_dseqs++;

   
    set_new_act_dseq(cube_ind,S,false,UNRES);
    check_red_cube_in(cube_ind,min_depth_lit,min_depth,true);    
           
    Red_cubes.push_back(cube_ind);
  }   

  
} /* end of function confirm_active_cubes */


/*===================================================

  F O R M _ L E F T _ R E D _ C U B E S

  This functions  updates Left_seqs[cube_ind]
  and  forms the set 'Left_red_cubes[curr_depth]'
  of cubes that switch their state from
  redundant to non-redundant.

  ASSUMPTIONS:
  1) There are no cubes redundant
  due to conflict
  2) Every non-satisfied redundant cube  has
  a corresponding active sequent
  3) 'Non_sgnl_cubes' is empty

  ======================================================*/
void ExQuant::form_left_red_cubes(Node &n,CUBE &Non_sgnl_cubes)
{


  vector <CubeMap> *pM;
  vector <int> &Red_cubes  = Left_red_cubes[curr_depth];
  Red_cubes.clear();

  if (Var_state[n.var-1].val == 0) pM = &Dseqs_to_check0;
  else pM = &Dseqs_to_check1;

  int x_ind = n.var - 1;

  CubeMap &Cands = (*pM)[x_ind];
  CubeMap::const_iterator i;

  for (i=Cands.begin(); i != Cands.end(); ++i)    {
    int cube_ind = i->first; 
    assert(Cube_info[cube_ind].state != ACTIVE);

    if (Cube_info[cube_ind].signal == 0)       {
      Non_sgnl_cubes.push_back(cube_ind);
      continue;
    }
    
    if (!contains_var(Act_dseqs[cube_ind],n.var))       {
      p();
      printf("sequent of cube %d is in Seqs_to_check[%d]\n",cube_ind,n.var-1);
      printf("but D-sequent: "); 
      print_dseq2(Act_dseqs[cube_ind],cube_ind);
      printf("does not depend on variable %d\n",n.var);
      printf("gcount = %ld\n",gcount);
      exit(ERROR);

    }  
 
    Red_cubes.push_back(cube_ind);  
    Left_dseqs[cube_ind].push_back(Act_dseqs[cube_ind]); 
    Sat_vars[cube_ind].push_back(n.var);
   
  }

} /* end of function form_left_red_cubes */

/*================================

  F L I P _ B R A N C H

  This function is called when
  the left branch of node 'n'
  has been explored
  =================================*/
void ExQuant::flip_branch(Node &n)
{
  if (dec_making_flag) {
    if (Var_state[n.var-1].xvar)    {
      printf("flip_branch: for node x%d at depth %d ",n.var,curr_depth);
      if (Var_state[n.var-1].val == 1) 
	printf("from assignment x%d to ~x%d\n",n.var,n.var); 
      else 
	printf("from assignment ~x%d to x%d\n",n.var,n.var);
    }
    else { // Y-variable
      printf("flip_branch: for node y%d at depth %d ",n.var,curr_depth);
      if (Var_state[n.var-1].val == 1) 
	printf("from assignment y%d to ~y%d\n",n.var,n.var); 
      else
	printf("from assignment ~y%d to y%d\n",n.var,n.var);
    }
  }
 
  Formula_size[n.var-1] = F.size();
  unassign_var(n);
  n.branch = RIGHT;  
} /* end of function flip_branch */
