#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh" 
#include "einline.hh"

/*================================================

  C O N F I R M _ A C T I V E _ C U B E S 1


  The difference of this function from
  'confirm_active_cubes' is that it ignores
  the active cubes marked in a hash table

  ------------------
  This function update information regarding
  protecting cubes for the cubes listed 
  in Cubes.

  ASSUMPTIONS:

  1) Cubes of 'Cubes' are already marked as ACTIVE

  2) Red_cubes is empty
 
  3) Active cubes to ignored   are
  marked in 'hash_table_cubes1.table'

  ===============================================*/
void ExQuant::confirm_active_cubes1(CUBE &Cubes,CUBE &Red_cubes,int depth)
{
  
  CUBE &table = hash_table_cubes1.table;
  int &marker = hash_table_cubes1.marker;
 
  for (uint i=0; i < Cubes.size(); i++)   {
    int cube_ind = Cubes[i];
    if (Cube_info[cube_ind].state != ACTIVE) continue;
    if (Cube_info[cube_ind].signal == 0) continue;

    int min_depth;
    int min_depth_lit;
     
    if (table[cube_ind] == marker) continue;
    bool success = recover_protection(cube_ind,depth,min_depth_lit,min_depth);
      
      
    if (success) continue;

    CUBE S;
    	    

    seq_gen_red_cube1(S,min_depth_lit);
    num_frw_blocked_cube_dseqs++;

   
    set_new_act_dseq(cube_ind,S,false,UNRES);
    check_red_cube_in(cube_ind,min_depth_lit,min_depth,true);         
    Red_cubes.push_back(cube_ind);
  }   

  
} /* end of function confirm_active_cubes1 */


/*============================================


  F I N D _ N E W _ P R O T E C T I O N 1


  
  The difference of this function from
  'find_new_protection' is that it ignores
  the active cubes marked in a hash table

  This function tries to find a new protection
  for every cube listed in 'Ped[cube_ind1]'
  that is still active

  ASSUMPTIONS:

  1) Cube 'F[cube_ind1]' just became redundant

  2) The active cubes to be ignored are marked 
  in 'hash_table_cubes1'

  =============================================*/
void ExQuant::find_new_protection1(CUBE &Red_cubes,int cube_ind1)
{

  int cube_ind2,lit_ind2;
 
  CUBE &table = hash_table_cubes1.table;
  int &marker = hash_table_cubes1.marker;

  while (Ped[cube_ind1].size() > 0)   {
    extract_pair(cube_ind2,lit_ind2,cube_ind1);
    if (Cube_info[cube_ind2].state != ACTIVE) 
      continue;
         
    if (table[cube_ind2] == marker) continue; // skip active cubes marked to be excluded
    int lit = F[cube_ind2][lit_ind2];
    if (Var_state[abs(lit)-1].val < 2)      
      continue; // assigned literals cannot be protected
       
    if (!prot_media_cubes)
      if (Cube_info[cube_ind2].signal == 0) continue; // no need to protect a media cube

    if (already_protected(cube_ind2,lit_ind2))       
      continue; // in case 'Ped' contains duplicates
      
  
    bool success = protect_one_cube_lit(cube_ind2,lit_ind2);
    if (success) continue;
  
    CUBE S;
    seq_gen_red_cube1(S,lit);
    num_frw_blocked_cube_dseqs++;

    int L = max_cube_recovery_depth(S);
    set_new_act_dseq(cube_ind2,S,false,UNRES);
    check_red_cube_in(cube_ind2,lit,L,true);   
    Red_cubes.push_back(cube_ind2);
  }

 
} /* end of function find_new_protection1 */


/*===================================================

  C H E C K _ F O R _ N E W _ R E D _ C U B E S 1

  The difference of this function from
  'check_for_new_red_cubes' is that it ignores
  the active cubes marked in a hash table

  This function updates the status of
  cubes protected by cubes listed in 'Red_cubes'
  that just got redundant.

  If no new protected cube exists for a cube,
  the latter is added to 'Red_cubes'
  
  ====================================================*/
void ExQuant::check_for_new_red_cubes1(CUBE &Red_cubes)
{

  while (Red_cubes.size() > 0)    {
    int cube_ind = Red_cubes.back();
    Red_cubes.pop_back();
    if (!prot_media_cubes)
      if (Cube_info[cube_ind].signal == 0) continue;
    find_new_protection1(Red_cubes,cube_ind);
  }


} /* end of function check_for_new_red_cubes1 */


/*========================

  E S T I M A T E

  =======================*/
inline int ExQuant::estimate(CUBE &C)
{
  int Res = 0;

  for (uint i=0; i < C.size(); i++)
    if (C[i] < 0) 
      Res += Flits1[-C[i]-1].size();
    else 
      Res += Flits0[C[i]-1].size();

  return(Res);

} /* end of function estimate */


/*====================================


  S Q U E E Z E _ P E D

  This function tries to reduce the
  size of Ped[cube_ind] if it
  obviously contains duplicates

  (because its size exceeds the total
  number of cubes that can be resolved
  with F[cube_ind])

  =====================================*/
void ExQuant::squeeze_Ped(int cube_ind)
{

 
  CUBE &C = F[cube_ind];

  if ((int) Ped[cube_ind].size() < 2*estimate(C))
    return; // there may not be any duplicates

 
  CUBE &Pairs = Ped[cube_ind];

  
  hash_table_cubes.change_marker();
  hash_table_cubes.started_using();
  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;

  int shift = 0;
  for (uint i=0; i < Pairs.size(); i++,i++)  {     
    int cube_ind = Pairs[i];     
    if (table[cube_ind] == marker)	{
      shift++;shift++;
      continue;
    }
    table[cube_ind] = marker;
    if (shift == 0) continue;
    Pairs[i-shift] = Pairs[i];
    Pairs[i-shift+1] = Pairs[i+1];
  }

  hash_table_cubes.done_using();
  Pairs.resize(Pairs.size() - shift);
 

} /* end of function squeeze_Ped */
