#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*========================================

  I N I T I A L I Z E 

  This function is called once

  Assumptions:

  1) Variable 'num_vars' is set
  2) Variable 'quant_thresh' is set
  3) DHF formula 'F' is formed
  4) Array 'Yvars' is formed
  5) Variable 'verbose' is set
  6) Variable 'selector' is set


  =========================================*/
void ExQuant::initialize()
{


  // initializing hash tables
  

  hash_table_lits.ht_init(2*num_vars+1);
  hash_table_aux.ht_init(2*num_vars+1);
  hash_table_cubes.ht_init(2*F.size()+1);
  hash_table_cubes1.ht_init(2*F.size()+1);

  // setting state variables 

  alg_state = DECISION_MAKING;
  bcp_first_node = DEC_LEFT;


  curr_depth = -1;
  num_missing_vars = 0;
 
  init_ncubes = F.size();

  /*-----------------------------------
    Memory allocation of stack related 
    data structures
    ------------------------------------*/

  // allocate  (and initialize if necessary) arrays of type vector <vector <int>> 
  //indexed by variables (Flits0,Flits1, New_sat_cubes)

  for (int i=0; i < num_vars; i++)    {
    CUBE  Dummy;
    // Flits0.push_back(Dummy);
    // Flits1.push_back(Dummy);  
    New_sat_cubes.push_back(Dummy);
  }


  // initialize literal counts  
  // initialize_cube_vects(F,Flits0,Flits1);


  // allocate and initialize Lit_activity0,Lit_activity1

  Lit_activity0.assign(num_vars,0);
  Lit_activity1.assign(num_vars,0);

  // allocate (and initialize if necessary)  arrays of type vector <type> 
  // indexed by variables (Var_state,Var_sel_depth,Red_cube_depth)
  //                      
  // fill out the state of variables
  VarState Var;
  Var.xvar = 1;
  Var.val = 2;
  Var.in_array = 0;

  for (int i=0; i < num_vars; i++)
    Var_state.push_back(Var);

  form_Xvars_Yvars();

  Var_sel_depth.assign(num_vars,-1);
  Red_cube_depth.assign(F.size(),-1);

  // allocate (and initialize if necessary)  arrays of type vector <int> 
  // indexed by cubes (cube_state, inact_cube_depth)
  // 

  CubeInfo el;
  el.state = ACTIVE;
  el.type = INITIAL;
  el.signal = 0;
  el.depth = num_vars+1;
  Cube_info.assign(F.size(),el);

  
  num_ycubes=0;
  Clean.assign(F.size(),1);
  for (uint i=0; i < F.size(); i++) {
    Cube_info[i].type = cube_type_check(F[i]);
    if (Cube_info[i].type == Y_CUBE) {
      num_ycubes++;
    }
  }
  dirt_ycubes = num_ycubes;
  init_ycubes = num_ycubes;
  printf("the number of initial ycubes is  %d\n",init_ycubes);
 
  // allocate auxiliary arrays 
  // initialize Act_ses, Left_seqs, Sat_vars
  // 

  for (uint i=0; i < F.size(); i++)    {
    DNF Dummy;
    Left_dseqs.push_back(Dummy);
    CUBE Dm;
    Sat_vars.push_back(Dm);
    Act_dseqs.push_back(Dm);
  }


  // initialize Left_red_cubes, New_sgnl_cubes and Cause_lit
  
  for (int i=0; i < num_vars; i++)   {
    CUBE Dummy;
    Left_red_cubes.push_back(Dummy);
    New_sgnl_cubes.push_back(Dummy);
  }

  Cause_lit.assign(F.size(),0);


  // initialize Seqs_to_check0, Seqs_to_check1
  for (int i=0; i < num_vars; i++)    {
    CubeMap Dummy;
    Dseqs_to_check0.push_back(Dummy);
    Dseqs_to_check1.push_back(Dummy);
  }

  

  // initialize some  data structures related to signal cubes

 

  if (Init_sgnl_cubes.size() == 0)    {
    p();
    if (prepr_mode) printf("prepr_mode is on\n");
    exit(ERROR);
  }



  Act_sgnl_cubes_cnt0.assign(num_vars,-1);
  Act_sgnl_cubes_cnt1.assign(num_vars,-1);
  set_init_counts();
  num_assgn_xvars = 0;
  num_assgn_yvars = 0;
  
  num_sgnl_xvars = Sgnl_xvars.size();
  num_sgnl_yvars = Sgnl_yvars.size();

  // statistics 
  num_confls = 0;
  num_xvar_decs = 0;
  num_yvar_decs = 0;
  num_ress = 0;
  num_dseqs = 0;
  num_resL_dseqs = 0;
  num_resR_dseqs = 0;
  num_resS_dseqs = 0;
  yvar_mode = true;
  form_unit_ycube_queue();
  merging_nodes = 0;
  hybrid_nodes = 0;
  num_frw_blocked_cube_dseqs = 0;
  num_frw_blocked_cube_dseqs0 = 0;
  num_bkw_blocked_cube_dseqs = 0;
  num_left_s_seqs = 0;
  num_sc_dseqs = 0;
  num_fc_dseqs = 0;
 
  // allocate memory for Ping and Ped

  for (uint i=0; i < F.size(); i++) {
    CUBE Dummy;
    Ping.push_back(Dummy);
    Ping[i].assign(F[i].size(),-1);
    Ped.push_back(Dummy);
  }
  
  num_act_sgnl_cubes = Init_sgnl_cubes.size();
  curr_num_sgnl_cubes = num_act_sgnl_cubes;
  max_num_sgnl_cubes = num_act_sgnl_cubes;
  Formula_size.assign(num_vars,-1);
 

  num_res_xvars=0;
  num_res_yvars = 0;
  num_res_xvars1= 0;
  num_res_yvars1 = 0;
  if (extra_stat_flag)
    Num_of_dseqs.assign(F.size(),0);

  
 
  
} /* end initialize */

