#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"



/*====================================================

  J O I N _ D S E Q S 

  This function is called when
  algorithm is in the state BPE
    
  It resolves sequents of left
  and right branches depending
  on the branching variable of n

  ASSUMPTIONS: 
  1) 'n' is not a conflict node that is 
  either n.left_cube < 0 or  n.right_child_cube < 0
  or both conditions hold


  ====================================================*/
void ExQuant::join_dseqs(Node &n)
{
  assert ((n.left_ecube < 0) || (n.right_ecube < 0)); // no conflict cube is processed in the BPE state


  if (n.left_ecube >= 0) 
    join_unsat_sat(n);
  else if (n.right_ecube >= 0)     {
    CUBE Old_protecting_cubes;
    join_sat_unsat(n,Old_protecting_cubes);
    check_for_new_red_cubes(Old_protecting_cubes);
  }
  else // no empty cubes in left or right branches
    join_sat_sat(n);
     


} /* end of function join_dseqs */

/*===============================================

  J O I N _ U N S A T _ S A T
 
  This function is called when the left
  branch of node 'n' is assigned an empty cube 'C'.

  Let S* be the set of active sequents depending
  on variable n.var where every sequent S of S*
  specifies a redundant variable y.
  This function resolves  C with every sequent
  S of S* on variable x and makes the resolvent
  a new active y-sequent

  ================================================*/
void ExQuant::join_unsat_sat(Node &n)
{

 

  CUBE &C = F[n.left_ecube];


  CUBE Cube_inds;

  int left_lit; // literal of n.var in cube C
 
  hash_table_cubes.change_marker();
  hash_table_cubes.started_using();

  if (n.lval == 0)    {
    find_candidate_red_cubes(Cube_inds,Dseqs_to_check1[n.var-1]);
    left_lit = -n.var;
    mark_sat_cubes(Flits1[n.var-1]); // mark cubes satisfied in the left branch
  }
  else  // n.lval == 1
    {
      find_candidate_red_cubes(Cube_inds,Dseqs_to_check0[n.var-1]);
      left_lit = n.var;
      mark_sat_cubes(Flits0[n.var-1]); // mark cubes satisfied in the left branch
    }

    
  bool ok;
  CUBE &table = hash_table_cubes.table;
  int marker = hash_table_cubes.marker;
 
  bool int_var_flag=false;
  if (Var_state[n.var-1].xvar) int_var_flag = true;

  for (uint i=0; i < Cube_inds.size();i++ )    {
    int cube_ind =Cube_inds[i];
    if (Cube_info[cube_ind].signal == 0){ // 
      if (Cube_info[cube_ind].state == REDUNDANT) check_red_cube_out(cube_ind);
      set_correct_cube_state(cube_ind);	  
      continue;
    }

    CUBE &S = Act_dseqs[cube_ind];
    CUBE NewSeq;

    num_resL_dseqs++;

    if (table[cube_ind] == marker)	{         
      ok = quick_res(NewSeq,S,-left_lit);
      if (!ok)  {
        p();
        printf("failed to find literal %d in D-sequent of cube  number %d: ",-left_lit,cube_ind);
        print_dseq3(cube_ind);
        print_cube(cube_ind);
        print_Assgns_cut(S);
        printf("gcount = %ld\n",gcount);
        exit(ERROR);
      }
	  
      num_sc_dseqs++;
	  
      int L = max_cube_recovery_depth(NewSeq);  
      set_new_act_dseq(cube_ind,NewSeq,true,JOIN);
      check_red_cube_in(cube_ind,0,L,false);
      continue;
    }

    // S does not contain varaible 'n.var' 
    if (!contains_lit(S,-left_lit))
      {
        p();
        printf("active sequent for cube %d: ",cube_ind); cout << S << endl;
        printf("does not contain literal %d\n",-left_lit);
        int var_ind = abs(left_lit)-1;
        printf("Formula_size[%d]=%d\n",var_ind,Formula_size[var_ind]);
        print_cube_state(cube_ind);
        printf("gcount = %ld\n",gcount);
        exit(ERROR);
      }

  
    num_fc_dseqs++;    
    ok = resolve(NewSeq,C,S,left_lit);
      
    if (int_var_flag) num_res_xvars++;
    else num_res_yvars++;
    assert(ok);	      
     

    int L = max_cube_recovery_depth(NewSeq);
    set_new_act_dseq(cube_ind,NewSeq,true,UNS_JOIN); 
    check_red_cube_in(cube_ind,0,L,false);
     
  }


  // build D-sequents for cubes satisfied by the right assignment to 'n.var'
 
  proc_cubes_sat_rbranch(n); // r5es_seqs.cc
  hash_table_cubes.done_using();

  
} /* end of function join_unsat_sat */

