#include <iostream>
#include <list>
#include <vector>
#include <set>
#include <map>
#include <string>
#include <queue>
#include <algorithm>
//
#include <stdio.h>
#include <assert.h>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
using namespace std;


#include "e0g-pqe+.hh"

//
//  PROVE_RED
//
bool ExQuant::prove_red(int cube_ind)
{

  CLAUSE C = F[cube_ind];
  assert(Curr_prim_scubes.size() == 0);
  Curr_prim_scubes.insert(cube_ind);

  DNF DD,BB;
  Msolver *Aux_slv = build_form(BB,DD);
  int prev_size = 1;
  
  while (true) {
    bool ok = check_time();
    if (!ok) {     
      return(false);
    }

    if (prev_size < (int) Curr_prim_scubes.size()) {
      delete Aux_slv;
      Aux_slv = build_form(BB,DD);
      prev_size = Curr_prim_scubes.size();
    }
    bool sat1 = Aux_slv->solve();
    if (!sat1) {
      delete Aux_slv;
      return(true);
    }

   
    MvecLits Assmps1;
    form_assumps1(Assmps1,Aux_slv);
   
    bool sat2 = Slv->solve(Assmps1);
    if (!sat2) { // removable boundary point
      rem_bnd_pnts++;
      CLAUSE D;
      gen_impl_cls(D,Assmps1);
      accept_new_clause(Slv,D);
      accept_new_clause(Aux_slv,D);
      Sol.push_back(D);
      DD.push_back(D);
      if (verbose) {
        printf("adding a solution cube: ");
        print_cube1(D);
      }
      new_ycubes++;
      if (max_new_ycubes > 0)
        if (new_ycubes >= max_new_ycubes)
          return(false);
      
      if (D.size() == 0) return(true);
      continue;
    }

    //
    // unremovable boundary point
    //
    unrem_bnd_pnts++;
    
    CLAUSE B;
    bool finished = blocking_clause(B,cube_ind);
    if (!finished) return(false);
    if (max_gcount > 0)
      if (gcount > max_gcount)
        return(false);
 
    if (B.size() == 0) {
      delete Aux_slv;
      return(true);
    }
   
    accept_new_clause(Aux_slv,B);
    BB.push_back(B);
  }

} /* prove_red */


//
// MANDATORY
//
bool ExQuant::mandatory(CUBE &A,int assgn,SCUBE &Ya,int cls_ind)
{

  CLAUSE &C = F[cls_ind];

  
  for (uint i=0; i < C.size(); i++) {
    if (C[i] == assgn) continue;
    int var_ind = abs(C[i])-1;
    if (A[var_ind] == -C[i]) continue; // skip falsif. assgn
    if (Xvar[var_ind]) return(false); // sat by an x-assgn
    if (Ya.find(A[var_ind]) != Ya.end())
      return(false); // sat by an y-assgn of Ya
  }

  return(true);
} /* mandatory */

//
//  REDUCE_Y_ASSGNS
//
void ExQuant::reduce_y_assgns(SCUBE &Ya,CUBE &A)
{

  for (uint i=0; i < A.size(); i++) {
    if (Xvar[i]) continue;
    CUBE *pClauses;
    if (A[i] < 0) pClauses = &Flits0[i];
    else pClauses = &Flits1[i];

    bool crucial = false;
    for (uint j=0; j < pClauses->size(); j++) {
      int cls_ind = (*pClauses)[j];
      if (mandatory(A,A[i],Ya,cls_ind)) {
        crucial = true;
        break;
      }
    }

    if (crucial) Ya.insert(A[i]);
  }

} /* reduce_y_assgns */


//
//  EXTR_Y_ASSGNS
//
void ExQuant::extr_y_assgns(CUBE &Ya,Msolver *Aux_slv)
{
  MboolVec &S = Aux_slv->model;
  for (int i=0; i < num_vars; i++) {
    if (Xvar[i]) continue;
    if (S[i] == Mtrue) Ya.push_back(i+1);
    else Ya.push_back(-(i+1));
  }
} /* extr_y_assgns */


//
//  FORM_SAT_ASSGN
//
void ExQuant::form_sat_assgn(CUBE &A,Msolver *Solver)
{
  MboolVec &S = Solver->model;
  for (int i=0; i < num_vars; i++) 
    if (S[i] == Mtrue) A.push_back(i+1);
    else A.push_back(-(i+1));
} /* form_sat_assgn */


//
//  FORM_ASSUMPS1
//
void ExQuant::form_assumps1(MvecLits &Assmps,Msolver *Aux_slv)
{

  CUBE Ya;

  extr_y_assgns(Ya,Aux_slv);
  add_assumps(Assmps,Ya);

} /* form_assumps1 */

//
//  BUILD_FORM2
//
Msolver* ExQuant::build_form(DNF &BB,DNF &DD)
{

  Msolver *Slvr = init_sat_solver(num_vars+Curr_prim_scubes.size());
  for (uint i=0; i < F.size(); i++) {
    if (Curr_prim_scubes.find(i) != Curr_prim_scubes.end())
      continue;
    if (Cube_info[i].state == REDUNDANT) {
      assert(Red_cube_depth[i] < 0);
      continue;
    }
    accept_new_clause(Slvr,F[i]);
  }

  DNF H;
  form_add_clauses(H);
  accept_new_clauses(Slvr,H);
  accept_new_clauses(Slvr,BB);
  accept_new_clauses(Slvr,DD);
  return(Slvr);
} /* build_form */
