#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh"
#include "einline.hh"

/*===================================

       R E S O L V E

  This function resolve cubes
  C1 and C2 on variable 'var(lit)'
  and put the result to cube
  'Res'

  ASSUMPTIONS:
  1) C1 and C2 are orhogonal only in one
  varaible
  2) C1 has literal 'lit' and
  C2 has literal '~lit'
  3) Cube 'Res' is empty
  4) C1 has no two or more literals
  of the same variable. 
  5) The same applies to C2
   
  ===================================*/
bool ExQuant::resolve(CUBE &Res,CUBE &C1,CUBE &C2,int lit)
{


  hash_table_lits.change_marker();
  hash_table_lits.started_using();
  CUBE &table = hash_table_lits.table;
  int marker = hash_table_lits.marker;

  // copy the literals of C1
  // but literal 'lit'
  
  bool found = false;
  for (uint i=0; i < C1.size(); i++)  {
    if (C1[i] == lit) {
      found = true;
      continue;
    }
    Res.push_back(C1[i]);
    if (C1[i] < 0) table[-C1[i]-1] = marker;
    else table[num_vars+ C1[i]-1] = marker;
  }

  if (!found)   {
    p();
    cout << "resolving cube C1: " << C1 << " with cube C2: " << C2;
    printf("on variable %d\n",abs(lit));
    printf("literal %d is not in C1\n",lit);
    printf("gcount = %ld\n",gcount);
    printf("F.size() = %d, init_ncubes = %d\n",(int) F.size(),init_ncubes);
    exit(ERROR);
  }
  
  // copy unique literals of C2
  // but literal '~lit'

  found = false;
  for (uint i=0; i < C2.size(); i++) {
    if (C2[i] == -lit) {
      found = true;
      continue;
    }
    if (C2[i] < 0) {
      if (table[-C2[i]-1] != marker) 
        Res.push_back(C2[i]);           
    }
    else // C2[i] > 0
      if (table[num_vars+C2[i]-1] != marker)
        Res.push_back(C2[i]);
  }
 
  hash_table_lits.done_using();
  return(true);
}  /* end of function resolve */
