#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh" 
#include "einline.hh"

/*==============================================

  S E T _ N E W _ A C T _ D S E Q

  This function makes 'S' a new sequent for
  the cube 'F[cube_ind]'

  ASSUMPTIONS:

  1) F[cube_ind] is an X-cube

  ================================================*/
void ExQuant::set_new_act_dseq(int cube_ind,CUBE &S,bool replace,DseqType type)
{


  //  if (curr_depth >= 0)  print_dseq2(S,cube_ind); 
  assert((curr_depth == -1) || (Cube_info[cube_ind].signal == 1)); // March 7

  CUBE &oldS = Act_dseqs[cube_ind];


  if (replace) {
    Node &n = Tree_path[curr_depth];
    assert(n.branch == RIGHT); // we replace sequents only in the right branches
    int lit;
    if (n.lval == 0) lit = n.var;
    else lit = -n.var;
    update_lit_presence(cube_ind,lit,oldS,S);
  }
  else 
    add_lit_presence(cube_ind,S);
    
   
  Act_dseqs[cube_ind] = S;
 
  if (num_dseqs++ % 100000 == 0)    {
    my_printf((char *) "number of D-sequents = %m\n",num_dseqs);
    my_printf((char *) "   number of ycubes = %m\n",num_ycubes);
    fflush(stdout);
  }

  if ((yvar_dec_mode == YVAR_LIT_ACTIVITY3) || (yvar_dec_mode == YVAR_LIT_ACTIVITY5))    {
    if (selector == 0)
      update_lit_activity(S);
    else 
      update_lit_activity(S,cube_ind);
    if (num_dseqs % DECAY_RATE == 0)
      decrease_lit_activity();	 
  }

  if (extra_stat_flag)   {
    if (cube_ind == wtch_cube_ind) {
      Dseqs_wtch_cube.push_back(S);
      Gcnts.push_back(gcount);
      Dseq_types.push_back(type);
    }
  }

} /* end of function set_new_act_dseq */



/*========================================

  U P D A T E _ L I T _ A C T I V I T Y

  =========================================*/
void ExQuant::update_lit_activity(CUBE &S,int cube_ind)
{

  int sgnl_depth = Cube_info[cube_ind].depth;

  for (uint i=0; i < S.size(); i++)    {
    int var_ind = abs(S[i])-1;
    if (Var_sel_depth[var_ind] < sgnl_depth) continue;
    if (S[i] > 0)  Lit_activity1[var_ind]++;
    else Lit_activity0[var_ind]++;
  }

} /* end of function update_lit_activity */

/*========================================

  U P D A T E _ L I T _ A C T I V I T Y

  =========================================*/
void ExQuant::update_lit_activity(CUBE &S)
{

  for (uint i=0; i < S.size(); i++)    {
    int var_ind = abs(S[i])-1;
    if (S[i] > 0)  Lit_activity1[var_ind]++;
    else Lit_activity0[var_ind]++;
  }

} /* end of function update_lit_activity */

/*============================================================

  M A X _ C U B E _ R E C O V E R Y _ D E P T H

  Given a redundant cube F[cube_ind], this function returns
  the maximum depth up to which this cube remains
  redundant.

  ASSUMPTIONS:
  1) F[cube_ind] is redundant 
  2) 'S' is the active D-sequent of F[cube_ind]
 
    
  ============================================================*/
int ExQuant::max_cube_recovery_depth(CUBE &S)
{
  int max_depth = -1;

  for (uint i=0; i < S.size(); i++)    {
    int var_ind = abs(S[i])-1;
    if (Var_sel_depth[var_ind] > max_depth)
      max_depth = Var_sel_depth[var_ind];
  }

  
  return(max_depth);
} /* end of function max_cube_recovery_depth */

/*=============================================

  D E C R E A S E _ L I T _ A C T I V I T Y

  ============================================*/
void ExQuant::decrease_lit_activity()
{
  for (int i=0; i < num_vars; i++)   {
    Lit_activity0[i] >>= 1;
    Lit_activity1[i] >>= 1;
  }

} /* end of function decrease_lit_activity */


/*=========================================

  R E M O V E _ L I T _ P R E S E N C E

  This function updates class members
  'Seqs_to_check0' and 'Seqs_to_check1'
  after cube 'F[cube_ind]' becomes
  non-redundant

  ===========================================*/
void ExQuant::remove_lit_presence(int cube_ind)
{

	
  CUBE &S = Act_dseqs[cube_ind];

  for (uint i=0; i < S.size(); i++) {
    int lit = S[i];
    if (lit < 0)
      Dseqs_to_check0[-lit-1].erase(cube_ind);
    else // S[i] > 0
      Dseqs_to_check1[lit-1].erase(cube_ind);
  }

} /* end of function remove_lit_presence */
 


/*=================================================

  U P D A T E _ L I T _ P R E S E N C E

  This function updates class members
  'Seqs_to_check0' and 'Seqs_to_check1'
  when cube 'F[cube_ind]' remains redundant
  but the old active sequent 'oldS'
  is replaced with a new one

  ASSUMPTIONS: 
  1) Sequent 'newS' is obtained by resolving cube 
  'oldS' with some other cube on variable 'abs(lit)'.

  2) 'oldS' contains 'lit' and 'newS' does not

  ====================================================*/
void ExQuant::update_lit_presence(int cube_ind,int old_seq_lit,
                                  CUBE &oldS,CUBE &newS)
{
   

  // Change the record of sequents having literal 'lit'
  if (old_seq_lit < 0)
    Dseqs_to_check0[-old_seq_lit-1].erase(cube_ind);
  else // old_seq_lit > 0
    Dseqs_to_check1[old_seq_lit-1].erase(cube_ind);

  // Change the record of sequents having literals
  // of 'newS' that are not in 'oldS'

  // mark variables of oldS
  hash_table_lits.change_marker();
  hash_table_lits.started_using();
  CUBE &table = hash_table_lits.table;
  int &marker = hash_table_lits.marker;

  for (uint i=0; i < oldS.size(); i++)
    table[abs(oldS[i])-1] = marker;

  for (uint i=0; i < newS.size(); i++) {
    int lit = newS[i];
    int var_ind0 = abs(lit)-1;
    if (table[var_ind0] == marker) continue;
    if (lit < 0)  Dseqs_to_check0[var_ind0][cube_ind] = 1;
    else  Dseqs_to_check1[var_ind0][cube_ind] = 1;
  }

  hash_table_lits.done_using();

} /* end of function update_lit_presence */

/*=========================================

  A D D _ L I T _ P R E S E N C E

  This function updates class members
  'Seqs_to_check0' and 'Seqs_to_check1'
  to take into account adding new sequent S
  for cube 'F[cube_ind]'

  ===========================================*/
void ExQuant::add_lit_presence(int cube_ind,CUBE &S)
{
  
  for (uint i=0; i < S.size(); i++) {
    int index = abs(S[i]) - 1;
    if (S[i] < 0) {
      if (Dseqs_to_check0[index].find(cube_ind) == Dseqs_to_check0[index].end())
        Dseqs_to_check0[index][cube_ind] = 1;
    }
    else     {
      if (Dseqs_to_check1[index].find(cube_ind) == Dseqs_to_check1[index].end())
        Dseqs_to_check1[index][cube_ind] = 1;
    }       
  }

} /* end of function add_lit_presence */

