#include <iostream>
#include <list>
#include <vector>
#include <queue>
#include <set>
#include <map>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <algorithm>
#include <ctime>
#include <climits>
#include <cassert>
#include <string.h>

#include "Solver.h"
#include "SimpSolver.h"
#include "System.h"
using namespace std;
#include "e0g-pqe+.hh" 
#include "einline.hh" 


/*=========================================

  R E M _ S G N L _ Y C U B E S


  Assumptions:
  1) Array 'Yvars' has
  been formed

  2) The value of num_vars 
  is correctly formed

  =======================================*/
void ExQuant::rem_sgnl_ycubes()
{


  CCUBE Buff;

  Buff.assign(num_vars,0);

  for (uint i=0; i < Yvars.size(); i++)
    Buff[Yvars[i]-1] = 1;

  assert(Init_sgnl_cubes.size() >  0);
 
  int count = 0;
  for (uint i=0; i < Init_sgnl_cubes.size(); i++)  {
    int cube_ind = Init_sgnl_cubes[i];
    char found = 0;

    CUBE &C = F[cube_ind];   
    for (uint j=0; j < C.size(); j++)  {
      if (Buff[abs(C[j])-1] == 1) continue;
      found = 1;
      break;
    }
    
    if (!found)  {
      count++;
      continue;
    }
    
    if (count > 0)
      Init_sgnl_cubes[i-count] = Init_sgnl_cubes[i];

  }

  Init_sgnl_cubes.resize(Init_sgnl_cubes.size() - count);


}  /* end of function rem_sgnl_ycubes*/

/*===============================================

  S E T _ C O R R E C T _ C U B E _ S T A T E

  This function is called in the BPE state
  when F[cube_ind] is in state REDUNDANT but does
  not have the signal status. So it does not make
  sense to make the current D-sequent of 
  F[cube_ind] independent of the branching
  variable.

  This function sets F[cube_ind] into state 
  SATISFIED if it is satisfied by an existing
  assignment or state ACTIVE

  ================================================*/
void ExQuant::set_correct_cube_state(int cube_ind)
{


  int smallest_depth= num_vars+1;
  int smallest_depth_ind=-1;

  CUBE &C = F[cube_ind];

  for (uint i=0; i < C.size(); i++)  {
    int var_ind = abs(C[i])-1;
    if (Var_state[var_ind].val == 2) continue;
    int val = (C[i] > 0)?1:0;
    if (val != value(var_ind)) // satisfying assignment
      if (smallest_depth > Var_sel_depth[var_ind]) {
        smallest_depth = Var_sel_depth[var_ind];
        smallest_depth_ind = i;
      }
  }

  if (smallest_depth_ind == -1) { // C is not satisfied
    Cube_info[cube_ind].state = ACTIVE;
    return;
  }

  Cause_lit[cube_ind] = -C[smallest_depth_ind];
  assert(smallest_depth >= 0);
  New_sat_cubes[smallest_depth].push_back(cube_ind);

} /* end of function set_correct_cube_state */
